;===============================================================================
;+
; Name:
;      NIS_MRGDATA
; Purpose:
;      Merge continguous/overlapping windows in a CDS/NIS raster
; Calling sequence:
;      qlm=NIS_MRGDATA(ql [ , /NOCHECK ] )
; Input:
;      ql : CDS QLDS (Quick Look Data Structure)
; Keyword parameters:
;      NOCHECK= (0/1) check whether the input structure is a valid QLDS
; Output:
;      qlm: new QLDS structure 
; Common blocks:
;      None
; Calls:
;      QLMGR,GT_NUMWIN,GT_DETECTOR,GT_DIMENSION,COPY_QLDS, 
;      ST_WINDATA,REP_TAG_VALUE,GT_WINDATA
; Description:
;      This routine looks in a QLDS for windows that overlap or that are 
;      continguous. It then creates a new QLDS structure, with windows of 
;      variable size (mostly in the wavelength/detector X direction) to cover 
;      the detector area of the overlapping windows. Isolated windows are 
;      left unaltered.
; Side effects:
;      In some instances the Y-size of some windows is increased too, due 
;      to a different offset in the overlapping windows (<- spectral slant).
;      The Y-size of all other windows is adjusted accordingly. For the 
;      future, it might be worth considering letting the Y-size vary as 
;      well (perhaps adding a PADDING keyword too). 
; Restrictions:
;      - Only works on NIS data with 3 (WAVELNTH,SOLAR_X,SOLAR_Y) or 4 
;      dimensions (WAVELNTH,SOLAR_X,SOLAR_Y,DEL_TIME).
;      - Assumes that pixels in the common areas of overlapping windows have 
;      the same values (but allowing for data drop-outs in one of the 
;      windows). If data have been processed with, e.g., VDS_DEBIAS, VDS_CALIB 
;      or NIS_CALIB (keyword ANGSTROM=0), that should be strictly true; the 
;      use of VDS_ROTATE and such, on the other hand, obviously invalidates 
;      this assumption. 
;      Also, if data have been binned, there might be problems with offsets 
;      between windows of sub-(binned-)pixel resolution. Example: a 35-pixels 
;      wide window starts at (X,Y)=(524,576), while the next one, because of 
;      the spectral slant on the detector, starts at (558,577). The last 
;      column of the first window overlaps with the first column of the 
;      second, but offset of 1 pixel in the Y direction on the detector. 
;      If data are binned in pairs in the Y direction, the bottom-right binned 
;      data element in the first window is the sum of detector pixels 
;      (524,576) and (524,577), while the corresponding (bottom-left) binned 
;      data element in the second window is the sum of pixels (524,577) 
;      and (524,578).
; Modification history:
;      V. Andretta (15/Dec/1997): Written; based on older routine MRG_NISDATA
;      V. Andretta (19/Jan/1998): Take into account binning (BINX/Y);
;                                 Allows for time sequences.
; v1.0 V. Andretta (19/Jan/1998): Allows for header-only QLDS (read with 
;                                 /HEADER keyword from READCDSFITS)
; v2.0 V. Andretta (26/Oct/1998): Eliminated duplicated call to COPY_QLDS
; v3.0 V. Andretta (23/Aug/2000): Fixed some problems appearing with binned 
;                                 data.
;-
;===============================================================================

  function NIS_MRGDATA,ql $
                      ,NOCHECK=nocheck


;===============================================================================
;Initial setup

  on_error,2
  if !debug ne 0 then on_error,0
  errmsg=''
  qlm=0

;===============================================================================
;Check input

  if n_elements(ql) eq 0 then begin
    errmsg='First parameter must be a Quick Look Data Structure (QLDS)'
    goto,handle_error
  endif

  if not keyword_set(nocheck) then begin
    QLMGR,ql,valid
    if not valid then begin
      errmsg='First parameter must be a valid Quick Look Data Structure (QLDS)'
      goto,handle_error
    end
  end


;===============================================================================
;Initial definitions

;Create new structure (returned anyway, even if no changes are made)

  qlm=COPY_QLDS(ql)

;Which detector?

  detector=GT_DETECTOR(ql)
  if detector ne 'NIS' then begin
    errmsg='Use with NIS data only'
    goto,handle_error
  endif

;What is the dimensionality of this data set?

  ql_dim=GT_DIMENSION(ql)
  Ndims=ql_dim.Ndims
;return if not at least 3 dimensions are present
  if Ndims ne 3 and Ndims ne 4 then begin
    print,'%I> NIS_MRGDATA: Data must have 3 or 4 dimensions. No changes made.'
    return,qlm
  endif

;Data actually loaded?

  if min(ql.detdesc.ixstop(0)) lt 0 then begin
    errmsg='Data not loaded'
    goto,handle_error
  endif

;How many windows?

  Nwin=GT_NUMWIN(ql)

;Find where wavelength- and Y-axis are

;Detector X (wavelength)
  index_xd=where(ql.detdesc(0).axes eq 'WAVELNTH')
  index_xd=index_xd(0)
  if index_xd lt 0 then begin
    errmsg='Wavelength axis not found'
    goto,handle_error
  endif
;Detector Y (solar Y)
  index_yd=where(ql.detdesc(0).axes eq 'SOLAR_Y')
  index_yd=index_yd(0)
  if index_yd lt 0 then begin
    errmsg='Y axis not found'
    goto,handle_error
  endif

;New storage?

  new_storage=(where(tag_names(ql) eq 'DETDATA'))(0)
  new_storage=new_storage lt 0


;===============================================================================
;Get the boundaries of the windows on the detector


  detx=ql.detdesc.detx
  binx=ql.detdesc.binx>1
 ;*** N.B.: Nxd is the number of pixels in the X direction spanned by the 
 ;*** data window
  Nxd=(abs(ql.detdesc.ixstop(index_xd))-ql.detdesc.ixstart(index_xd)+1)*binx

  dety=ql.detdesc.dety
  biny=ql.detdesc.biny>1
 ;*** N.B.: see above
  Nyd=(abs(ql.detdesc.ixstop(index_yd))-ql.detdesc.ixstart(index_yd)+1)*biny


;===============================================================================
;Find which windows overlap or are continguous

  iwlist=intarr(Nwin)-1
  Nwinm=Nwin*0
  for iw=0,Nwin-1 do begin
    iwrange=where(abs(detx-detx(iw)) le Nxd(iw) and $
                  abs(dety-dety(iw)) le Nyd(iw))
    iwinm=max(iwlist(iwrange))
    if iwlist(iwrange(0)) lt 0 then begin
      iwlist(iwrange)=Nwinm
      Nwinm=Nwinm+1
    endif else begin
      iwlist(iwrange)=iwinm
    endelse
  endfor
;return if the number of windows will remain the same (i.e.: no overlapping 
;windows)
  if Nwinm eq Nwin then begin
    print,'%I> NIS_MRGDATA: No windows overlap. No changes made.'
    return,qlm
  endif


;===============================================================================
;Create new data description

 ;print,'%I> NIS_MRGDATA: Updating data description...'
  mdesc=replicate(ql.detdesc(0),Nwinm)
  ixstart_xd=0
  for iwm=0,Nwinm-1 do begin
    iwrange=where(iwlist eq iwm)
    Nwrange=n_elements(iwrange)
    mdesc(iwm)=ql.detdesc(iwrange(0))
;define the new dimension of the window
    xd0m=min(detx(iwrange))
    xd1m=max(detx(iwrange)+Nxd(iwrange)-1)
    yd0m=min(dety(iwrange))
    yd1m=max(dety(iwrange)+Nyd(iwrange)-1)
    binxm=mdesc(iwm).binx>1
    binym=mdesc(iwm).biny>1
    Nxdm=xd1m-xd0m+1
    Nydm=yd1m-yd0m+1
;put these informations in the data description structure
    mdesc(iwm).ixstart(index_xd)=ixstart_xd
    mdesc(iwm).ixstop(index_xd) =ixstart_xd+Nxdm/binxm-1
    mdesc(iwm).ixstop(index_yd) =           Nydm/binym-1
    for ii=1,n_elements(iwrange)-1 do $
      mdesc(iwm).label=mdesc(iwm).label+'+'+ql.detdesc(iwrange(ii)).label
    mdesc(iwm).detx=xd0m
    mdesc(iwm).dety=yd0m
   ;mdesc(iwm).wavelength=total(ql.detdesc(iwrange).wavelength)/Nwrange
   ;mdesc(iwm).wavemin=min(ql.detdesc(iwrange).wavemin)
   ;mdesc(iwm).wavemax=max(ql.detdesc(iwrange).wavemax)
    ixstart_xd=ixstart_xd+Nxdm/binxm
;create a new handle
    if new_storage then mdesc(iwm).handle=handle_create()
  endfor
;In order to avoid troubles, force all the dimensions (other than 
;wavelength) to be equal; data will be padded later
  mdesc.ixstop(index_yd)=max(mdesc.ixstop(index_yd))


;===============================================================================
;Create new structure

  print,'%I> NIS_MRGDATA: Creating new structure...'

;Free some space by eliminating the data in the new structure 
;(these data are already available from the original structure anyway)

  if new_storage then $
    for iw=0,Nwin-1 do ST_WINDATA,qlm,iw,0 $
  else $
    qlm=REP_TAG_VALUE(qlm,0,'DETDATA')

;Update data description structure in the new QLDS

  qlm=REP_TAG_VALUE(qlm,mdesc,'DETDESC')


;===============================================================================
;Merge data

  print,'%I> NIS_MRGDATA: Rearranging data...'

;Go through all the 'new' windows

  for iwm=0,Nwinm-1 do begin

;Some definitions

    iwrange=where(iwlist eq iwm)
    Nwrange=n_elements(iwrange)
    mdesc=qlm.detdesc(iwm)
    detxm=mdesc.detx
    binxm=mdesc.binx>1
    detym=mdesc.dety
    binym=mdesc.biny>1
    dims=abs(mdesc.ixstop)-mdesc.ixstart+1
    Nxdm=dims(index_xd)*binxm           ; X-size of new detector window 
    Nydm=dims(index_yd)*binym           ; Y-size of new detector window
    missing=mdesc.missing

;Create new data array... and store it if new_storage eq 1

    if Ndims eq 3 then $
      mdata=fltarr(dims(0),dims(1),dims(2),/NOZERO) $
    else $
      mdata=fltarr(dims(0),dims(1),dims(2),dims(3),/NOZERO)
    mdata(*)=missing
    mdata0=mdata     ; needed to initialize data from each window
    xdm=detxm+indgen(dims(index_xd))*binxm ; detector X
    ydm=detym+indgen(dims(index_yd))*binym ; detector Y

;Copy data from all overlapping windows

    for irange=0,Nwrange-1 do begin
      mdata_cur=mdata0
      iw=iwrange(irange)
      dix=(xdm-detx(iw))/binx(iw) & Nix=Nxd(iw)/binx(iw)
      xdrange=where(dix ge 0 and dix lt Nix)
      diy=(ydm-dety(iw))/biny(iw) & Niy=Nyd(iw)/biny(iw)
      ydrange=where(diy ge 0 and diy lt Niy)
      if dix(0)*binx(iw) ne xdm(0)-detx(iw) then $
        print,'%W> NIS_MRGDATA: Rounded off X position of window '+TRIM(iw) $
             +' (BINX='+TRIM(binx(iw))+')'
      if diy(0)*biny(iw) ne ydm(0)-dety(iw) then $
        print,'%W> NIS_MRGDATA: Rounded off Y position of window '+TRIM(iw) $
             +' (BINY='+TRIM(biny(iw))+')'
      data=GT_WINDATA(ql,iw)
      mdata_cur(xdrange,*,ydrange,*)=temporary(data)
;copy only those pixels that are missing or are in areas not covered by 
;the preceding windows:
      goodpix=where(mdata_cur ne missing and mdata eq missing)
      if goodpix(0) ge 0 then mdata(goodpix)=mdata_cur(goodpix)
    endfor

;final touches to the description of the data (only if the current window 
;is the result of a merger)

    if Nwrange gt 1 then begin
      goodpix=where(mdata ne missing)
      if goodpix(0) ge 0 then begin
        qlm.detdesc(iwm).min=min(mdata(goodpix))
        qlm.detdesc(iwm).max=max(mdata(goodpix))
      endif
      qlm.detdesc(iwm).origin(index_xd)= $
        min(ql.detdesc(iwrange).origin(index_xd))
      qlm.detdesc(iwm).wavelength=qlm.detdesc(iwm).origin(index_xd)+ $
        qlm.detdesc(iwm).spacing(index_xd)*0.5*(Nxdm-1)
      qlm.detdesc(iwm).wavemin=min(ql.detdesc(iwrange).wavemin)
      qlm.detdesc(iwm).wavemax=max(ql.detdesc(iwrange).wavemax)
    endif

;Store data (new storage) or extend data array (old storage)

    if new_storage then $
     ;qlm.detdesc(iwm).handle=handle_create(value=mdata,/no_copy) $
      ST_WINDATA,qlm,iwm,mdata,/no_copy $
    else $
      if iwm eq 0 then mdata_full=mdata else mdata_full=[mdata_full,mdata]

  endfor

;If QLDS employes old storage method:

  if not new_storage then begin
    print,'%I> NIS_MRGDATA: Replacing data...'
    qlm=REP_TAG_VALUE(qlm,mdata_full,'DETDATA')
  endif


;===============================================================================
;Register new structure with QLMGR

  qlm.ql_id=0
  QLMGR,qlm,valid
  if not valid then $
;Something went wrong...
    print,'%W> NIS_MRGDATA: New data structure is not a valid QLDS'

;===============================================================================
;Normal end

  return,qlm


;===============================================================================
;Error handling

HANDLE_ERROR:
  if n_elements(errmsg) eq 0 then errmsg=''
  errmsg='%E> NIS_MRGDATA: '+errmsg
  print,errmsg
  return,qlm

end
